#include <fstream>
#include <iostream>
#include <queue>
#include <stack>
#include <string>
#include <unordered_set>

const std::string ALPHABET = "abcdefghijklmnopqrstuvwxyz";

const char *const FORMAT_UNDERSCORE = "\033[4m";
const char *const END_FORMATTING = "\033[0m";
const char *const DICTIONARY_PATH = "./dictionary.txt";

std::string input(const char *message) {
    if (message != nullptr)
        std::cout << message;

    std::cout << FORMAT_UNDERSCORE;
    std::flush(std::cout);

    std::string ret;
    std::getline(std::cin, ret);
    std::cout << END_FORMATTING;
    return ret;
}

std::string fileRead(const std::string &filePath) {
    std::ifstream file(filePath, std::ios::binary);
    if (!file.is_open())
        throw std::runtime_error("File couldn't be found");

    std::string content;
    const int BUF_LENGTH = 1024;
    char buf[BUF_LENGTH];
    while (file.read(buf, BUF_LENGTH))
        content.append(buf, file.gcount());
    content.append(buf, file.gcount());

    if (!file.eof() && file.fail())
        throw std::runtime_error("Error: Failed to read the file '" + filePath +
                                 "'");

    return content;
}

std::unordered_set<std::string> readDictionary() {
    std::string fileContent = fileRead(DICTIONARY_PATH);
    std::unordered_set<std::string> set;
    size_t last = 0, idx;
    while ((idx = fileContent.find('\n', last)) != std::string::npos) {
        std::string sub = fileContent.substr(last, idx - last);
        set.insert(sub);
        last = idx + 1;
    }

    set.insert(fileContent.substr(last));
    return set;
}

std::stack<std::string>
wordChain(std::string &w1, std::string &w2,
          const std::unordered_set<std::string> &dictionary) {
    std::unordered_set<std::string> visited;
    std::queue<std::stack<std::string>> queue;
    {
        std::stack<std::string> first;
        first.push(w1);
        queue.push(first);
    }
    visited.insert(w1);

    while (!queue.empty()) {
        std::stack<std::string> front = queue.front();
        queue.pop();
        std::string latestWord = front.top();
        if (latestWord == w2)
            return front;
        for (int i = 0; i < latestWord.size(); i++) {
            for (char c : ALPHABET) {
                std::string nextWord = latestWord;
                nextWord[i] = c;
                bool isVisited = visited.find(nextWord) != visited.end();
                if (isVisited)
                    continue;
                bool isValidWord = dictionary.find(nextWord) != dictionary.end();
                if (!isValidWord)
                    continue;
                std::stack<std::string> nextStack = front;
                nextStack.push(nextWord);
                queue.push(nextStack);
                visited.insert(nextWord);
            }
        }
    }

    return std::stack<std::string>();
}

int main() {
    std::cout << "Welcome to TDDD86 Word Chain.\n"
              << "If you give me two English words, I will transform the\n"
              << "first into the second by changing one letter at a time.\n\n"
              << "Please type two words: ";
    std::flush(std::cout);

    std::unordered_set<std::string> dictionary = readDictionary();

    std::string userInput = input(nullptr);
    size_t wordSep = userInput.find(' ');
    std::string wordOne = userInput.substr(0, wordSep);
    std::string wordTwo = userInput.substr(wordSep + 1);

    // std::string wordOne = "lead", wordTwo = "gold";
    std::cout << "Chain from " << wordTwo << " to " << wordOne << ':'
              << std::endl;
    std::stack<std::string> resolutionWords =
        wordChain(wordOne, wordTwo, dictionary);

    bool isFirst = true;
    if (!resolutionWords.empty()) {
        do {
            std::string resolutionWord = resolutionWords.top();
            resolutionWords.pop();
            if (isFirst) {
                isFirst = false;
            } else {
                std::cout << ", ";
            }
            std::cout << resolutionWord;
        } while (!resolutionWords.empty());
    } else
        std::cout << "Couldn't find any resolution :(";
    std::cout << std::endl;

    std::cout << "Have a nice day!" << std::endl;
    return 0;
}
